//
// カメラ計測基板用スケッチ EOS-kiss shutter control
// EOS00 LAN接続タイプカメラモジュールのひな型 by T.Fukatsu
// PoEでFSから電源制御されることを前提の動作
//
// NTPアクセス後、丁度良い時間になるまでスリープ(選択制、NTP失敗時はスリープなし)
// 指定した時間帯で、AA(59)分間:電源ON(シャッター自動撮影), BB(01)分間:電源off
// 電源on時はWebサーバon、Webサーバでもカメラ電源、撮影ボタンを制御
// Twitterはテスト用に1ループ(AA分+BB分)毎にツイート。AA分+BB分は60で割れる数
// SWが解放時はテストモード(NTPアクセス後、すぐに起動、カメラ電源は常時on)
// SWが短絡時は別のAA,BB分を使用(単独のインターバル動作用)
// 引数 relay2(カメラの電源)=on/off, capture=start, begin/end=??, reset=on
//
// 2014/4/1 ver.1.0  EOS_module02をベースに作成
// 2014/4/4 ver.1.01 起動/スリープ時刻を任意に設定可能に
// 2014/5/26 ver.1.02 割り込み・NTPコメントアウト。常に常時モード
// 2014/11/20 ver.1.04 スタンドアロンテスト用
// 2014/12/1 ver.1.05 EOS電源起動時の電圧降下対策
// 2014/12/15 ver.1.06 Eye-Fi6分間隔でresetするためEOSはon/off。DCコンバータ変更。Etherリセット位置変更
// 2015/6/5 ver.1.061 initial_sleep*1000UL-INITIAL_OFFSET がマイナスにならないif文修正
// 2015/11/24 ver.1.062 SW解放(通常)時をテストモード(すぐに起動、起動時間59分) 11/30修正
//
#include <SPI.h>
#include <Ethernet.h>
#include <MsTimer2.h>
#include <Twitter.h>
#include <Time.h>
#include <EthernetUdp.h>

PROGMEM prog_char version_name[] = "[cam_EOS00 ver1.062]";
PROGMEM prog_char web_header1[] = "HTTP/1.0 200 OK\nServer: arduino\nCache-Control: no-store, n";
PROGMEM prog_char web_header2[] = "o-cache, must-revalidate\nPragma: no-cache\nConnection: clos";
PROGMEM prog_char web_header3[] = "e\nContent-Type: text/html\n";
PROGMEM prog_char web_top1[]    = "<html><head><title>Camera Control Web Server</title></";
PROGMEM prog_char web_top2[]    = "head><body><h3>Camera Control ver.1.062</h3><hr/><form>";
PROGMEM prog_char web_top3[]    = "<table border>";
PROGMEM prog_char web_end[]     = "</table></form><HR/>2014.4.4 by T.FUKATSU(NARO)</body></html>";

// 動作時間に関係するパラメータ
const unsigned long device_Wait = 1000UL*2UL; //(ms) 動作安定のためのウェイト
const unsigned long ether_Wait = 1000UL*40UL; //(ms) LAN関係が安定するまでのウェイト
const unsigned long pint_Wait = 1000UL*6UL; //(ms)カメラが立ちあがる時間
unsigned long wake_Period = 0; //(ms) MODE_pinによって wake_Period1 or wakePeriod2
unsigned long sleep_Period = 0; //(ms) MODE_pinによって sleep_Period1 or sleep_Period2
const unsigned long wake_Period1 = 1000UL*60UL*59UL; //(ms) カメラが起動している時間
const unsigned long sleep_Period1 = 1000UL*60UL*1UL; //(ms) 電源off時間 (wake+sleepは60で割れるように)
const unsigned long wake_Period2 = 1000UL*60UL*6UL; //(ms)
const unsigned long sleep_Period2 = 1000UL*60UL*54UL; //(ms)

// Arduinoのpin設定
const byte shutter_pin = 2;      //camera shutter on
const byte half_shutter_pin = 3; //camera focus on
const byte cam_relay = 4;       //EOS-kiss camera power Relay, Lで電源on, Hで電源off(HRD05003用)
const byte MODE_pin = 5;        //H:Period2 <--> L:Period1
const byte LED_pin1 = 6;        //output LED, 起動時onでNTPアクセス成功時off
const byte LED_pin2 = 7;      //output LED, これがonの時はSWがonでPerion2を選択
const byte reset_pin = 8;      //for Ethernet Shield reset(起動時にhighにする必要有)
const byte ether_relay = 9;    // ether 5Vの電源on/off

// インターネット接続関係 PRGMEMで書くと動作せず...
EthernetServer server(80);
byte mac[] = { 0x90, 0xA2, 0xDA, 0x0E, 0xF2, 0x0C };
byte  ip[] = { 192, 168, 20, 100 };
// byte dns_ip[] = { 192, 168, 100, 1 }; //dnsという名前だとエラー
// byte gateway[] = { 192, 168, 100, 1 }; //最後が*.1でない場合は設定必要
// byte subnet[] = { 255, 255, 255, 0 };

// NTPサーバアクセス関係
const unsigned int localPort = 8888;
IPAddress timeServer(133, 243, 238, 164); //ntp.nict.jp
//IPAddress timeServer(150, 26, 2, 66); //stm1.cc.affrc.go.jp(maffin内部のみ)
const byte NTP_PACKET_SIZE = 48;
byte packetBuffer[NTP_PACKET_SIZE];
EthernetUDP Udp;

//Twitterのトークン
//Twitter twitter("1968299359-7iVIEr7K48cnsybLOhNYAc6g5F0SesnBabil2Dx"); // fs-classic00
Twitter twitter("2516131285-FShhGMtPuirWEOk7pDq5CjC2CgQCLFbH0mL4zNC"); // fs-eos00

//データ取得時の変数
unsigned int raw[6]; // A/Dで得られた生電圧を保存(0-1023)
char str[64]; // コメント表示,Twitterなどテンポラリ文字列を格納
char msg[141]; // Twitteする本文を格納,最大140文字
String arg_data=""; //Webサーバアクセス時の引数を格納

// 制御用フラグ
boolean flag = true; //これがfalseになるとloop()は終了し、再度loop()を実行
boolean mode_flag = false; //SWがoff状態(Period1動作)はfalse, trueにするとPeriod2(SWがon状態)
boolean reset_flag = false; //trueのときsoftReset. Web上でreset=onとしたとき用
boolean relay2_flag = false; //Web表示用、カメラの電源状態
// ユーザ設定フラグ
boolean twitter_flag = true; //twitteしないときはfalse(所内ネットからはtwitte不可)
boolean ntp_flag = true; // NTPで時刻同期取得を行うかどうか。
boolean init_wait_flag = false; //Period1の時、NTP時刻同期成功後に初回動作を切りの良い時間まで待つかどうか

// NTPおよび内部時刻保存用
int NTP_time = -1; //(s) from 00 min.からの経過秒数 (求められなかった時は-1)
const int INITIAL_OFFSET = 0; //(ms) NTPアクセス後、メインループに入るまでのウェイトオフセット  
int default_offset = 0;       //(ms) 割り込み毎のオフセット
const int DEFAULT_OFFSET1 = 0; //(ms) 割り込み毎のオフセット(Period1用)
const int DEFAULT_OFFSET2 = 0; //(ms) 割り込み時のオフセット(Period2用)
const byte DEFAULT_RESET_HOUR = 5; //(0-23)　この時間になる直前の起動終了後にソフトリセット
byte offline_hour = DEFAULT_RESET_HOUR; //(0-23) NTPにアクセス出来なかったときはこの初期値から内部時計スタート
byte offline_min = 0; //(0-59)

//日中・夜間動作切替用
byte dayhour_begin = 0; //dayhour_startからdayhour_endまでを日中と判断
byte dayhour_end = 24; //日中の間は一定時間毎に自動でシャッター

//
// セットアップ
//
void setup(){ 
  // initial  
  Serial.begin(9600);
  strcpy_P(str, (char*)version_name);
  Serial.println(str);
  // pin setup
  pinMode (shutter_pin, OUTPUT);
  pinMode (half_shutter_pin, OUTPUT);
  pinMode (cam_relay, OUTPUT);
  pinMode (ether_relay, OUTPUT);
  pinMode (LED_pin1, OUTPUT);
  pinMode (LED_pin2, OUTPUT);
  pinMode (reset_pin, OUTPUT);
  pinMode (MODE_pin, INPUT);
 
  // power on
  digitalWrite(cam_relay, HIGH); //カメラ最初は電源OFF
  relay2_flag = false;
  digitalWrite(reset_pin, LOW);
  digitalWrite(LED_pin1, HIGH); //NTPアクセスLED
  digitalWrite(ether_relay, HIGH); //Etherは電源ON 
  digitalWrite(half_shutter_pin, LOW); //シャッターはまだ押さない
  digitalWrite(shutter_pin, LOW); //シャッターはまだ押さない
 
  if( digitalRead(MODE_pin) == HIGH ){   
    mode_flag = true; //MODE_pin(SW)が短絡してHIGHの時はPeriod2を使用
    digitalWrite(LED_pin2, HIGH); //チェック用LEDを点灯
    wake_Period = wake_Period2;
    sleep_Period = sleep_Period2;
    default_offset = DEFAULT_OFFSET2;
  } else {
    mode_flag = false;
    digitalWrite(LED_pin2, LOW); //ソフトリセット時前の状態が残るため
    wake_Period = wake_Period1;
    sleep_Period = sleep_Period1;
    default_offset = DEFAULT_OFFSET1;
  }
  
//  delay(device_Wait); //不要かも？ reset_pinをhighにするまでにコマンドがあるので大丈夫？
  digitalWrite(reset_pin, HIGH); //ここでEtherシールドをリセットしないとハング
  analogReference(INTERNAL); //アナログ入力は0-1Vに切り替え

  // network
  Serial.println(F("waiting..."));
  delay(ether_Wait);
  Ethernet.begin(mac, ip);
  // Ethernet.begin(mac, ip, dns_ip, gateway, subnet); //フルに記述する場合

  // NTPで時刻取得
  int initial_sleep = 0; //(sec)
  if( ntp_flag ){
    Udp.begin(localPort);
    delay(device_Wait);
    sendNTPpacket(timeServer); //AccessNTP内でも実行しているから不要かも？
    delay(device_Wait); //NTPサーバにうまくアクセスするため必要
   //NTP access
    AccessNTP();
    Udp.stop();
    if( NTP_time != -1 ){
      digitalWrite(LED_pin1, LOW); //チェック用LEDを消灯
      int one_loop_sec = (wake_Period+sleep_Period)/1000; //(sec)
      initial_sleep = one_loop_sec - (NTP_time % one_loop_sec); //(sec)
      offline_hour = hour();
      offline_min = (NTP_time + initial_sleep)/60; //(min)
//      offline_min = (NTP_time - (NTP_time % one_loop_sec) + one_loop_sec)/60; //(min)
      if( offline_min >= 60 ){
        offline_min = 0;
        offline_hour = offline_hour + 1;
      }
      if ( offline_hour >= 24 ) offline_hour = 0;
      if(initial_sleep <= INITIAL_OFFSET/1000) initial_sleep=initial_sleep+one_loop_sec;
      Serial.print(F("initial wait:"));
      Serial.print(initial_sleep);
      Serial.println(F("(s)"));
      //きちんとNTPで時刻補正できたかをTwitte
      strcpy_P(msg, (char*)version_name);
      sprintf(str, "%d : %d//%d (s)  interval:%d(s)",hour(), NTP_time, initial_sleep, one_loop_sec);
      strcat(msg, str);
      Serial.println(msg);
      if(twitter_flag) ArduinoTwitte(msg); 
    } else {
      // NTPにアクセスできなかった場合   
      initial_sleep = sleep_Period/1000; //(sec)
      offline_hour = DEFAULT_RESET_HOUR;
      offline_min = 0; //(sec)
      // if( !mode_flag ) initial_sleep = 0; //Period1&NTP失敗時は初期ウェイトなし --> init_wait_flagに吸収
    }
  }   
  
  // (NTPにアクセスできてそうでなくとも)通常モードの時は規定時間になるまでスリープ
  if( mode_flag ){
    digitalWrite(ether_relay, LOW); //Etherは電源OFF
    digitalWrite(reset_pin, LOW);  
    digitalWrite(cam_relay, HIGH);
    relay2_flag = false;
  }
  if( !init_wait_flag && !mode_flag ) initial_sleep = 0;
  if(initial_sleep*1000UL > INITIAL_OFFSET) delay(initial_sleep*1000UL-INITIAL_OFFSET);
 
  // 割り込みの設定, 通常モードは指定した起動時間経過毎に割り込みで時刻チェック
  MsTimer2::set(wake_Period-default_offset, Interrupt_function);
  MsTimer2::start();    
}


//
// 割り込みルーチン, 基本的にはスリープ
//
void  Interrupt_function(){
    //テストモードの時は割り込み無視。
    interrupts(); //delayを使用するためには必要
    MsTimer2::stop(); //割り込みが終わるまで次の割り込みまでのカウントを停止
    flag = false;
    
    // ローカル時刻更新
    int one_loop_min = (wake_Period+sleep_Period)/1000/60; //(min)
    offline_min = offline_min + one_loop_min;
    if( offline_min >= 60 ){
      offline_min = offline_min - 60;
      offline_hour = offline_hour + 1;
    }
    if( offline_hour >= 24 ) offline_hour = 0;

    // １日1回リセット
    if( offline_hour == DEFAULT_RESET_HOUR  && offline_min < one_loop_min){
      digitalWrite(cam_relay, HIGH);
//      delay( sleep_Period );
      digitalWrite(ether_relay, LOW); //Etherは電源OFF
      digitalWrite(reset_pin, LOW);      
      softReset();
    }
    
    //各種電源off
    digitalWrite(half_shutter_pin, LOW);
    digitalWrite(shutter_pin, LOW);
    if ( mode_flag ){
      digitalWrite(ether_relay, LOW); //Etherは電源OFF
      digitalWrite(reset_pin, LOW);     
    }
    digitalWrite(cam_relay, HIGH);
    relay2_flag = false; 
    delay( sleep_Period );
    Serial.println(F("back to loop"));
    MsTimer2::start();
}


//
// 画像撮影
//
void capture(){
  Serial.println(F("Capture"));
  digitalWrite(half_shutter_pin, HIGH);
  delay(pint_Wait);
  digitalWrite(shutter_pin, HIGH);
  delay(device_Wait);
  digitalWrite(half_shutter_pin, LOW);
  digitalWrite(shutter_pin, LOW);
}

//
// メインループ
//
void loop(){

  // 初期設定
  flag = true;
  reset_flag = false;
  sprintf(msg,""); // 念のため初期化
  sprintf(str,""); // 念のため初期化

  digitalWrite(half_shutter_pin, LOW); 
  digitalWrite(shutter_pin, LOW);   

   digitalWrite(cam_relay, LOW); //カメラの電源ON
   relay2_flag = true;
   delay(device_Wait);  
   digitalWrite(reset_pin, LOW);  // Ether電源常時Onでも時々ハングするのでリセット準備

  // power on  
  if( mode_flag ){ 
    digitalWrite(ether_relay, HIGH); //Etherは電源ON
    delay(device_Wait);
    digitalWrite(reset_pin, HIGH);
    Serial.println(F("waiting..."));
    delay(ether_Wait);
    Ethernet.begin(mac, ip);
  } else {
    delay(device_Wait);
    digitalWrite(reset_pin, HIGH);
    Ethernet.begin(mac, ip);    
    delay(pint_Wait);
  }
  
  // カメラ撮影
  if( dayhour_begin <= offline_hour && dayhour_end > offline_hour ){
    capture();
    Serial.println(F("data sending..."));
  }


  //Twitteに書き込む内容を作成
  strcpy_P(msg, (char*)version_name);
  sprintf(str, "[%lu] %d(h):%d(m) ", millis()/1000, offline_hour, offline_min);
  strcat(msg,str);
  ReadSensor();
  sprintf(str ,"ad0:%d, ad1:%d, ad2:%d, ad3:%d, ad4:%d, ad5:%d",
                raw[0], raw[1], raw[2], raw[3], raw[4], raw[5]); 
  strcat(msg, str);
  Serial.println(msg);
  if(twitter_flag) ArduinoTwitte(msg);
  

  // Web server start
  server.begin();
  Serial.println(F("Web Server start"));
  while(flag){
      EthernetClient client = server.available();
      arg_data = ""; //Webサーバアクセス時の引数を初期化
      boolean arg_flag = true; //HTTPのヘッダを読み込み終えるまでtrue
      if (client && flag) {
        // an http request ends with a blank line
         boolean currentLineIsBlank = true;
         while (client.connected() && flag) {
            delay(1); //timing delay else webserver hangs.
            if (client.available() && flag) {
              char c = client.read();
              if( arg_flag ) arg_data.concat(c);
              if (c == '\n' && currentLineIsBlank) {
                //引数を確認して動作
                //カメラの電源制御
                if(arg_data.indexOf("relay2=off") != -1 ){
digitalWrite(cam_relay, HIGH); //カメラの電源OFF
relay2_flag = false;                 
                }
                if(arg_data.indexOf("relay2=on") != -1 ){
digitalWrite(cam_relay, LOW); //カメラの電源ON
relay2_flag = true;
                }
                //機器のソフトウェアリセット
                if(arg_data.indexOf("reset=on") != -1 ){
                  reset_flag = true;
                }
                //Web上からマニュアル撮影
                if(arg_data.indexOf("capture=start") != -1 ){
    capture();         
                }
/*                
                int numind = -1; 
                char numstr[2];
                //起動パラメータ変更
                if((numind = arg_data.indexOf("begin=")) != -1 ){               
                  numstr[0] = arg_data.charAt(numind+6);
                  numstr[1] = arg_data.charAt(numind+7);                 
                  dayhour_begin = atoi(numstr);        
                }
               if((numind = arg_data.indexOf("end=")) != -1 ){               
                  numstr[0] = arg_data.charAt(numind+4);
                  numstr[1] = arg_data.charAt(numind+5);                 
                  dayhour_end = atoi(numstr);        
                }
*/                
                // Webレスポンス
                strcpy_P(str, (char*)web_header1);
                client.print(str);
                strcpy_P(str, (char*)web_header2);
                client.print(str);
                 strcpy_P(str, (char*)web_header3);
                client.println(str);               
                strcpy_P(str, (char*)web_top1);
                client.print(str);
                strcpy_P(str, (char*)web_top2);
                client.print(str);
                strcpy_P(str, (char*)web_top3);
                client.println(str);               
                sprintf(msg, "<tr><td>");
                strcpy_P(str, (char*)version_name);             
                strcat(msg,str);
                unsigned long sec = millis()/1000;                
                sprintf(str, "[%lu sec]: </td></tr>", sec);
                strcat(msg,str);
                client.println(msg);               
                ReadSensor();             
                for (int analogChannel = 0; analogChannel < 6; analogChannel++) {
                  client.print(F("<tr><td>ad"));
                  client.print(analogChannel);
                  client.print(F(":"));
                  client.print(raw[analogChannel]);
                  client.println(F("<br /></td></tr>"));
                }
                client.print(F("<tr><td>&nbsp;</td></tr>"));
                client.print(F("<tr><td>cam: <input type=\"submit\" name=\"relay2\" value=\"on\">"));
                if( relay2_flag ){
                  client.print(F("on"));
                } else {
                  client.print(F("off"));
                }
                client.print(F("<input type=\"submit\" name=\"relay2\" value=\"off\"></td></tr>"));

                client.print(F("<tr><td>capture: <input type=\"submit\" name=\"capture\" value=\"start\">"));
                client.print(F("</td></tr>"));

               client.print(F("<tr><td>work hour:"));
               client.print(dayhour_begin);
               client.print(F("--"));
               client.print(dayhour_end); 
               client.print(F("h [now: "));
               client.print(offline_hour);
               client.print(F(":"));
               client.print(offline_min);
               client.print(F("]"));               
               client.print(F("</td></tr>"));
                
                strcpy_P(str, (char*)web_end);
                client.println(str);
                Serial.println(F("http accessed. "));
                break;
              }
              if (c == '\n') {
                // you're starting a new line
                arg_flag = false;
                currentLineIsBlank = true;
              } 
              else if (c != '\r') {
                // you've gotten a character on the current line
                currentLineIsBlank = false;
              }
           }
         }
         // give the web browser time to receive the data
         delay(1);
         // close the connection:
         client.stop();
         if( reset_flag ) softReset();       
      }
   }
   Serial.println(F("loop finish")); 
}

//
// NTPリクエストを送信
//
unsigned long sendNTPpacket(IPAddress& address)
{
   // set all bytes in the buffer to 0
   memset(packetBuffer, 0, NTP_PACKET_SIZE); 
   // Initialize values needed to form NTP request
   // (see URL above for details on the packets)
   packetBuffer[0] = 0b11100011;   // LI, Version, Mode
   packetBuffer[1] = 0;     // Stratum, or type of clock
   packetBuffer[2] = 6;     // Polling Interval
   packetBuffer[3] = 0xEC;  // Peer Clock Precision
   // 8 bytes of zero for Root Delay & Root Dispersion
   packetBuffer[12]  = 49; 
   packetBuffer[13]  = 0x4E;
   packetBuffer[14]  = 49;
   packetBuffer[15]  = 52;
   // all NTP fields have been given values, now
   // you can send a packet requesting a timestamp:         
   Udp.beginPacket(address, 123); //NTP requests are to port 123
   Udp.write(packetBuffer,NTP_PACKET_SIZE);
   Udp.endPacket(); 
}

//
// NTPで時刻を取得
//
void AccessNTP(){
  Serial.println(F("send NTP packet"));
  sendNTPpacket(timeServer);
  delay(1000); //念のため(信州大のコードを参照)
  if ( Udp.parsePacket() ) {
     Udp.read(packetBuffer, NTP_PACKET_SIZE);
     // 時刻情報はパケットの40バイト目からはじまる4バイトのデータ
     unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
     unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
     // NTPタイムスタンプは64ビットの符号無し固定小数点数
     //（整数部32ビット、小数部32ビット）
     // 1900年1月1日0時との相対的な差を秒単位で表している
     // 小数部は切り捨てて、秒を求めている
     unsigned long secsSince1900 = highWord << 16 | lowWord;
     // NTPタイムスタンプをUNIXタイムに変換する
     // UNITタイムは1970年1月1日0時からはじまる
     // 1900年から1970年の70年を秒で表すと2208988800秒になる
     const unsigned long seventyYears = 2208988800UL;
     // NTPタイムスタンプから70年分の秒を引くとUNIXタイムが得られる
     unsigned long epoch = secsSince1900 - seventyYears;
     // Timeライブラリに時間を設定(UNIXタイム)
     // 日本標準時にあわせるために＋9時間しておく
     setTime(epoch + (9 * 60 * 60));
//     sprintf(str, "JST is %d / %d / %d", year(), month(), day());
//     Serial.print(str);
     sprintf(str, " %d : %d : %d", hour(), minute(), second());
     Serial.println(str);
     NTP_time = second()+minute()*60; //00分からの秒数を算出
  } else {
    Serial.println(F("NTP access failed."));
    NTP_time = -1;
  } 
}

//
// A/D接続のセンサ値を読み込む(8回取りこんで平均する)
//
void ReadSensor(){
   for(byte analogChannel=0; analogChannel<6; analogChannel++) {
      analogRead(analogChannel); // dummy,チャンネル間の電圧吸い込み誤差を低減
      raw[analogChannel] = 0;
      for( byte i=0; i<8; i++){
         raw[analogChannel] += analogRead(analogChannel);
      }
      raw[analogChannel] = (raw[analogChannel]+4) >> 3; //切捨なので8/2を足してから8で割る
   }
}



//
// msgの内容をtwitteする
//
void ArduinoTwitte( char msg[141] )
{
  Serial.println(F("connecting ..."));
  if (twitter.post(msg)) {
    int status = twitter.wait();
    if (status == 200) {
        Serial.println(F("twitte OK"));
    } else {
      Serial.print(F("failed : code "));
      Serial.println(status);
    }
  } else {
    Serial.println(F("connection failed."));
  }
  Serial.println(F("twitte end"));
}


/*
//
// デバッグ用メモリチェック
//
void check_mem()
{
  uint8_t *stackptr, *heapptr;
  stackptr = (uint8_t *)malloc(4);    //use stackptr temporarily
  heapptr = stackptr;                 //save value of heap pointer
  free(stackptr);                     //set stackptr to 0
  stackptr = (uint8_t *)(SP);         //save value of stack pointer
  
  Serial.print(F("Stack ptr :"));
  Serial.print((long)stackptr);
  Serial.print(F("<--> Heap ptr :"));
  Serial.println((long)heapptr);     //stackptr - heapptrが残りメモリ量
}
*/

//
// ソフトウェアリセット 
//
void softReset()
{
  Serial.println("Reboot!");
  MsTimer2::stop(); //MsTimer2::start()のままだとハングする...
  delay(1000);
  asm volatile ("  jmp 0");  
}
